# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractQpsCounter, AbstractQpsTimer, AbstractQueryPerSecond
from .core._c import QpsCounter
from .core._t import QpsTimer
from .core._q import QueryPreSecond
from .core.api import create_qps



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from w8_auto_py.typings import Function
from w8_auto_py.common.callbacks import AbstractCallback, filter_function, filter_async_function
from w8_auto_py.common.validator import callable_validator, coroutine_func_validator


class Callback(AbstractCallback):

    def __init__(self, *callbacks):
        super().__init__(*filter_function(*callbacks))

    def add(self, callback: Function) -> None:
        callable_validator(
            error_message=f"{callback} 需要一个函数或对象"
        ).is_validate(callback)

        self.set.add(callback)

    def extend(self, *callbacks) -> None:
        return self.set.update(filter_function(*callbacks))


class ACallback(AbstractCallback):

    def __init__(self, *acallbacks):
        super().__init__(*filter_async_function(*acallbacks))

    def add(self, callback: Function) -> None:
        coroutine_func_validator(
            error_message=f"{callback} 需要一个协程函数"
        ).is_validate(callback)

        self.set.add(callback)

    def extend(self, *callbacks) -> None:
        return self.set.update(filter_async_function(*callbacks))



# @Time     : 2021/3/27
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import typing

try:
    import simplejson as json
except ImportError:
    import json

from w8_auto_py.typings import PyObject, PathType
from w8_auto_py.global_vars import G_ENCODING


class JsonUtil:

    @classmethod
    def to_json(cls,
                obj: PyObject,
                encoding: str = G_ENCODING,
                ensure_ascii: bool = False,
                **kwargs) -> str:
        """
        python object -> JSON
        :return:
        :param obj:
        :param encoding:
        :param ensure_ascii:
        :param kwargs:
        :return: str
        """
        try:
            return json.dumps(obj,
                              encoding=encoding,
                              ensure_ascii=ensure_ascii,
                              **kwargs)
        except Exception as e:
            raise ValueError(f"json 解析失败 {e}")

    @classmethod
    def to_dict(cls,
                json_string: str,
                encoding: str = G_ENCODING,
                **kwargs) -> typing.Dict[str, typing.Any]:
        """
        JSON -> python dict
        :param json_string:
        :param encoding:
        :param kwargs:
        :return: dict
        """
        if not json_string:
            return {}

        if isinstance(json_string, dict):
            return json_string

        try:
            return json.loads(json,
                              encoding=encoding,
                              **kwargs)

        except json.JSONDecodeError as e:
            raise ValueError(f"JSON to python dict error: {e}")

    @classmethod
    def to_file(cls,
                obj: PyObject,
                file: PathType,
                encoding: str = G_ENCODING,
                ensure_ascii: bool = False,
                indent: int = 4) -> None:
        """
        写入文件
        Args:
            obj:
            file:
            encoding:
            ensure_ascii:
            indent:

        Returns:

        """
        with open(file, mode="w", encoding=encoding) as f:
            json.dump(obj, f, ensure_ascii=ensure_ascii, indent=indent)



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractRunner, AbstractRunStrategy, IRunner, IRunStrategy
from .core._strategy import (
    CoroutineStrategy,
    CoroutineFunctionStrategy,
    FunctionStrategy,
    RunStrategy,
    create_runner_strategy
)
from .core._runner import AsyncRunner
from .core.api import runner_factory, create_arunner, run


# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import (
    AbstractEmitter,
    AbstractEmitterMaster,
    AbstractEmitterStrategy,
    AbstractEmitterTask,
    EmitterReturnType,
    EmitterMasterReturnType,
    IEmitter,
    IEmitterTask,
    delay
)
from .core._t import EmitterQueue, Done, join
from .core._master import EmitterGroup, EmitterCollection
from .core._strategy import (
    CoroutineEmitterStrategy,
    FunctionEmitterStrategy,
    EmitterStrategy,
    create_emitter_strategy
)
from .core._e import Emitter
from .core.api import (
    Emitters,
    EmitterFactory,
    create_emitter,
    create_emitter_group,
    create_emitter_collection,
    emitter_decorator,
    start
)



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from w8_auto_py.typings import FunctionIterable
from w8_auto_py.common.validator import is_function, is_coroutine_function


def filter_function(*functions) -> FunctionIterable:
    """
    过滤函数
    :param functions:
    :return:
    """
    return (
        func for func in functions if is_function(func)
    )


def filter_async_function(*afunctions) -> FunctionIterable:
    """
    过滤异步函数
    :param afunctions:
    :return:
    """
    return (
        afunc for afunc in afunctions if is_coroutine_function(afunc)
    )



# @Time     : 2021/3/27
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .global_vars import NumberUnit, TimeUnit, G_ENCODING
from .exceptions import (
    ValidateException,
    NotCallableError,
    NotCoroutineError,
    NotCoroutineFunctionError
)
from w8_auto_py import typings
from w8_auto_py import util
from .common import validator
from .common import callbacks
from .common import qps
from .common import runner
from .common import emitter



# @Time     : 2021/3/27
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import enum
import typing

EnumValue = typing.TypeVar("EnumValue")


class EnumUtil(object):

    @staticmethod
    def create_dict_by_enum(enums: enum.Enum) -> typing.Dict:
        """
        通过枚举创建字典
        :param enums:
        :return:
        """
        if not issubclass(enums, enum.Enum):
            raise TypeError(f"{enums} 需要一个枚举类")

        return {item: item.value for _, item in enumerate(enums)}

    @staticmethod
    def get_enum_key(mapping: dict, key: enum.Enum) -> EnumValue:
        """
        获取枚举 key
        :param mapping:
        :param key:
        :return:
        """
        if not isinstance(key, enum.Enum):
            raise TypeError(f"{key} 需要一个枚举对象")

        if not isinstance(mapping, dict):
            mapping = {}

        return mapping.get(key)

    @staticmethod
    def get_enum_value(enums: enum.Enum) -> EnumValue:
        """
        获取枚举值
        :param enums:
        :return:
        """
        if not isinstance(enums, enum.Enum):
            raise TypeError(f"{enums} 需要一个枚举对象")

        return enums.value



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractValidator
from .core._util import is_function, is_async_task, is_coroutine, is_coroutine_function
from .core._validators import (
    CallableValidator,
    CoroutineValidator,
    CoroutineFunctionValidator
)
from .core.api import (
    validator_factory,
    callable_validator,
    coroutine_validator,
    coroutine_func_validator
)


# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import abc
import typing

from w8_auto_py.typings import NumberTypes

T = typing.TypeVar("T")


class AbstractQpsTimer(metaclass=abc.ABCMeta):
    """
    抽象计时器
    """

    @property
    def is_timeout(self) -> bool:
        """
        是否超时
        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_current(self) -> NumberTypes:
        """
        获取当前时间
        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def set_timeout(self) -> T:
        """
        设置超时
        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def clear(self) -> T:
        """
        清除超时
        Returns:

        """
        raise NotImplementedError()


class AbstractQpsCounter(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def count(self) -> NumberTypes:
        raise NotImplementedError()

    @abc.abstractmethod
    def increment(self, num: NumberTypes) -> T:
        """
        增加计数
        Args:
            num:

        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def clear(self) -> T:
        """
        清零
        Returns:

        """
        raise NotImplementedError()


class AbstractQueryPerSecond(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def start(self) -> None:
        """
        开始
        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def clear(self) -> T:
        """
        清零
        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def auto_increment(self) -> T:
        """
        自增
        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> T:
        raise NotImplementedError()



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractCallback
from .core._util import filter_function, filter_async_function
from .core._callbacks import Callback, ACallback
from .core.api import callback_factory, callback, acallback


# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import asyncio
from collections.abc import Callable


def is_function(value) -> bool:
    return isinstance(value, Callable)


def is_coroutine(value) -> bool:
    return asyncio.iscoroutine(value)


def is_coroutine_function(value) -> bool:
    return asyncio.iscoroutinefunction(value)


def is_async_task(task) -> bool:
    return isinstance(task, asyncio.Task)



# @Time     : 2021/3/26
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
class ValidateException(Exception):
    pass


class NotCoroutineError(ValidateException):
    pass


class NotCoroutineFunctionError(ValidateException):
    pass


class NotCallableError(ValidateException):
    pass



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import abc


class AbstractValidator(metaclass=abc.ABCMeta):
    """
    抽象校验器
    """

    def __init__(self, error_message: str):
        self._error_message = error_message

    @property
    def error_message(self) -> str:
        if self._error_message is None:
            return ""
        return self._error_message

    @abc.abstractmethod
    def is_validate(self, value) -> bool:
        """
        校验方法
        :param value:
        :return:
        """
        raise NotImplementedError("NotImplementedError .is_validate(value)")

    def __call__(self, value, **kwargs) -> bool:
        return self.is_validate(value)



# @Time     : 2021/3/28
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import asyncio
import typing

from w8_auto_py.global_vars import TimeUnit
from w8_auto_py.typings import NumberTypes
from w8_auto_py.common.emitter import AbstractEmitterMaster, AbstractEmitter, IEmitterTask, join


class EmitterGroup(AbstractEmitterMaster):
    """
    并发组
    """

    def __init__(self,
                 task: IEmitterTask,
                 delay: NumberTypes,
                 time_unit: TimeUnit = TimeUnit.MS):
        """
        并发组管理器
        :param task:
        :param delay:
        :param time_unit:
        """
        super().__init__(task, delay, time_unit)
        self.__emitters: typing.List[AbstractEmitter] = []

    @property
    def empty(self) -> bool:
        return not self.length

    @property
    def emitters(self) -> typing.List[AbstractEmitter]:
        return self.__emitters

    @property
    def length(self) -> int:
        return len(self.emitters)

    def add(self, emitter: AbstractEmitter, **kwargs) -> None:
        """
        单个添加并发器
        Args:
            emitter:

        Returns:

        """
        if not self._is_emitter(emitter):
            raise TypeError("emitter 必须是一个并发对象")
        self.emitters.append(emitter)

    def extends(self, *emitters) -> None:
        """
        批量添加并发器
        Args:
            *emitters:

        Returns:

        """
        if not emitters:
            return

        self.emitters.extend((emitter for emitter in emitters if self._is_emitter(emitter)))

    async def run(self):
        for emitter in self:
            self.task.add_task(asyncio.create_task(emitter.start()))
            await asyncio.sleep(self.delay)

        await join(self.task)

    def _is_emitter(self, emitter) -> bool:
        return isinstance(emitter, AbstractEmitter)

    def __iter__(self) -> typing.Iterable[AbstractEmitter]:
        for _, item in enumerate(range(self.length)):
            """ 为空, 自动退出 """
            yield self._pop_head()

    def _pop_head(self) -> AbstractEmitter:
        return self.emitters.pop(0)


class EmitterCollection(AbstractEmitterMaster):
    """
    并发集合
    """

    def __init__(self,
                 task: IEmitterTask,
                 delay: NumberTypes = 500,
                 time_unit: TimeUnit = TimeUnit.MS):
        """
        并发集合管理器
        :param task:
        :param delay:
        :param time_unit:
        """
        super().__init__(task, 2 ** 0, time_unit)
        self.__emitters: typing.Set[typing.Union[AbstractEmitter,
                                                 EmitterGroup]] = set()

    @property
    def length(self) -> int:
        return len(self.emitters)

    @property
    def emitters(self) -> typing.Set[typing.Union[AbstractEmitter,
                                                  EmitterGroup]]:
        return self.__emitters

    def add(self, emitter: typing.Union[AbstractEmitter,
                                        EmitterGroup], **kwargs) -> None:
        """
        单个添加
        Args:
            emitter:

        Returns:

        """
        if not self._is_emitter(emitter):
            raise TypeError("emitter 必须是并发对象或并发组")

        self.emitters.add(emitter)

    def extends(self, *emitters) -> None:
        """
        批量添加并发器、并发组
        Args:
            *emitters:

        Returns:

        """
        self.emitters.update({emitter for emitter in emitters if self._is_emitter(emitter)})

    async def run(self):
        for emitter in self:
            self.task.add_task(asyncio.create_task(emitter.start()))
            await asyncio.sleep(self.delay)

        await join(self.task)

    def _is_emitter(self, emitter: typing.Union[AbstractEmitter,
                                                EmitterGroup]) -> bool:
        return isinstance(emitter, (AbstractEmitter, EmitterGroup))

    def __iter__(self) -> typing.Iterable[typing.Union[AbstractEmitter,
                                                       EmitterGroup]]:
        for _ in enumerate(range(self.length)):
            """ 为空, 自动退出 """
            yield self.emitters.pop()



# @Time     : 2021/3/26
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import enum

_G_BASE_NUMBER = 10
G_ENCODING: str = "utf-8"


class TimeUnit(enum.Enum):
    """
    时间单位
    """
    S = _G_BASE_NUMBER ** 0  # 秒
    MS = _G_BASE_NUMBER ** -3  # 毫秒
    US = _G_BASE_NUMBER ** -6  # 微秒


class NumberUnit(enum.Enum):
    """
    数学单位
    """
    B = _G_BASE_NUMBER ** 2  # 百
    Q = _G_BASE_NUMBER ** 3  # 千
    M = _G_BASE_NUMBER ** 6  # 兆



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import functools

from w8_auto_py.typings import ArgsType, Function, KwargsType, PoolExecutorTypes
from w8_auto_py.common.runner import AbstractRunner, AsyncRunner


def runner_factory(runner: AbstractRunner,
                   coro_or_func: Function,
                   args: ArgsType, **kwargs) -> AbstractRunner:
    if not issubclass(runner, AbstractRunner):
        raise TypeError("runner 必须是 AbstractRunner 子类")

    f = functools.partial(runner, coro_or_func=coro_or_func, args=args)

    return f(**kwargs)


def create_arunner(coro_or_func: Function,
                   args: ArgsType = None,
                   kwargs: KwargsType = None,
                   executor: PoolExecutorTypes = None) -> AsyncRunner:
    """
    异步 runner 工厂函数
    :param coro_or_func:
    :param args:
    :param kwargs:
    :param executor:
    :return:
    """
    return runner_factory(
        AsyncRunner,
        coro_or_func=coro_or_func,
        args=args,
        kwargs=kwargs,
        executor=executor
    )


def run(runner: AbstractRunner):
    return runner.run()



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import functools

from w8_auto_py.common.callbacks import AbstractCallback, Callback, ACallback


def callback_factory(callback: AbstractCallback, *functions) -> AbstractCallback:
    if not issubclass(callback, AbstractCallback):
        raise TypeError("callback 必须是 AbstractCallback 子类")

    f = functools.partial(callback)
    return f(*functions)


def callback(*functions) -> Callback:
    return callback_factory(Callback, *functions)


def acallback(*afunctions) -> ACallback:
    return callback_factory(ACallback, *afunctions)



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from w8_auto_py.typings import NumberTypes
from w8_auto_py.common.qps import AbstractQpsCounter


class QpsCounter(AbstractQpsCounter):

    def __init__(self, count: NumberTypes):
        self.__count = int(count)

    @property
    def count(self) -> NumberTypes:
        return self.__count

    def increment(self, num: NumberTypes) -> None:
        if not isinstance(num, (int, float)):
            raise TypeError("num must be int or float")
        self.__count += num

    def clear(self) -> None:
        self.__count = 0



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from w8_auto_py.typings import (
    ArgsType,
    Function,
    KwargsType,
    NumberTypes,
    PoolExecutorTypes
)
from w8_auto_py.global_vars import TimeUnit
from w8_auto_py.common.emitter import create_emitter_strategy, delay, AbstractEmitter, IEmitterTask, EmitterStrategy
from w8_auto_py.common.validator import is_function, is_coroutine_function


class Emitter(AbstractEmitter):
    """
    并发器
    """

    def __init__(self,
                 coro_or_func: Function,
                 task: IEmitterTask,
                 max_len: int,
                 delay: NumberTypes = 500,
                 time_unit: TimeUnit = TimeUnit.MS,
                 args: ArgsType = None,
                 executor: PoolExecutorTypes = None,
                 kwargs: KwargsType = None):
        """
        并发器对象
        :param coro_or_func:
        :param task:
        :param max_len:
        :param delay:
        :param time_unit:
        :param args:
        :param executor:
        :param kwargs:
        """
        self.__coro_or_func = coro_or_func
        self.__task = task
        self.__args = () if not isinstance(args, tuple) else args
        self.__kwargs = {} if not isinstance(kwargs, dict) else kwargs
        self.__executor = executor

        """ 辅助变量 """
        self.__delay = delay
        self.__max_len = max_len
        self.__time_unit = time_unit
        self.__cache_delay: NumberTypes = 0

    @property
    def coro_or_func(self) -> Function:
        return self.__coro_or_func

    @property
    def delay(self) -> NumberTypes:
        if not self.__cache_delay:
            self.__cache_delay = delay(self.__delay, self.__time_unit)
        return self.__cache_delay

    @property
    def max_len(self) -> int:
        """
        控制最大并发数量
        :return:
        """
        if not isinstance(self.__max_len, int):
            return 2 ** 31
        return abs(self.__max_len)

    @property
    def task(self) -> IEmitterTask:
        return self.__task

    @property
    def args(self) -> ArgsType:
        return self.__args

    @property
    def kwargs(self) -> KwargsType:
        return self.__kwargs

    @property
    def executor(self):
        return self.__executor

    async def _impl_run(self):
        if is_coroutine_function(self.coro_or_func):
            """ 协程函数策略 """
            strategy = create_emitter_strategy(
                EmitterStrategy.COROUTINE,
                self, args=self.args, kwargs=self.kwargs
            )
            return await strategy.run()

        elif is_function(self.coro_or_func):
            """ 普通函数策略 """
            strategy = create_emitter_strategy(
                EmitterStrategy.FUNCTION,
                self, args=self.args, executor=self.executor
            )
            return await strategy.run()

        else:
            raise ValueError("coro_or_func 参数需要一个协程函数或普通函数")

    async def run(self):
        return await self._impl_run()

    def __str__(self):
        return f"{self.__coro_or_func.__name__} + {self.delay}"



# @Time     : 2021/3/26
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import sys
import typing
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from pathlib import Path

"""
全局类型文件
"""

_AnyStringMapping = typing.Mapping[str, typing.AnyStr]
_AnyStringDict = typing.Dict[str, typing.AnyStr]

PyObject = typing.Union[
    _AnyStringMapping, _AnyStringDict,
    typing.List[_AnyStringMapping], typing.List[_AnyStringDict],
    typing.Tuple[_AnyStringMapping], typing.Tuple[_AnyStringDict]
]

Function = typing.Callable
FunctionMap = typing.Dict[typing.Any, Function]
FunctionIterable = typing.Iterable[Function]
FunctionList = typing.List[Function]
FunctionSet = typing.Set[Function]

ArgsType = typing.Tuple[typing.Any]
KwargsType = typing.Dict[str, typing.Any]
NumberTypes = typing.Union[int, float]

PoolExecutorTypes = typing.Union[
    ThreadPoolExecutor,
    ProcessPoolExecutor
]

PathType = typing.Union[str, Path]

_PY_3743 = (3, 7, 4, 3)
_PY_38 = (3, 8)
Interface = object

if sys.version_info >= _PY_38:
    Interface = typing.Protocol

else:
    try:
        import typing_extensions

        Interface = typing_extensions.Protocol
    except ImportError:
        pass



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import functools

from w8_auto_py.common.validator import AbstractValidator
from w8_auto_py.common.validator import (
    CallableValidator,
    CoroutineValidator,
    CoroutineFunctionValidator
)


def validator_factory(validator: AbstractValidator,
                      error_message: str,
                      **kwargs) -> AbstractValidator:
    """
    校验器工厂
    :param validator:
    :param error_message:
    :param kwargs:
    :return:
    """
    if not issubclass(validator, AbstractValidator):
        raise TypeError("validator 必须是 AbstractValidator 子类")
    f = functools.partial(validator, error_message=error_message)
    return f(**kwargs)


def callable_validator(error_message: str) -> CallableValidator:
    return validator_factory(CallableValidator, error_message)


def coroutine_validator(error_message: str) -> CoroutineValidator:
    return validator_factory(CoroutineValidator, error_message)


def coroutine_func_validator(error_message: str) -> CoroutineFunctionValidator:
    return validator_factory(CoroutineFunctionValidator, error_message)



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import enum

from w8_auto_py.typings import NumberTypes
from w8_auto_py.global_vars import TimeUnit
from w8_auto_py.common.qps import AbstractQueryPerSecond, QpsCounter, QpsTimer


class RunStates(enum.Enum):
    """
    运行状态
    """
    STOP, START, PAUSE = range(3)  # 0 stop, 1 start 2 pause


class QueryPreSecond(AbstractQueryPerSecond):

    def __init__(self, timeout: NumberTypes, date_unit: TimeUnit = TimeUnit.MS):
        """

        Args:
            timeout:        超时时间
            date_unit:      时间单位
        """

        self.__st: NumberTypes = 0  # 开始时间
        self.__states: RunStates = RunStates.STOP  # 运行状态
        self.__timeout = timeout  # 超时时间
        self._counter = QpsCounter(0)  # 计数器
        self._timer = QpsTimer(date_unit)  # 计时器

    @property
    def counter(self):
        return self._counter

    @property
    def timer(self):
        return self._timer

    @property
    def is_start(self) -> bool:
        """
        是否开始
        Returns:

        """
        return self.__states == RunStates.START

    def auto_increment(self) -> None:
        """
        自增
        Returns:

        """
        if not self.is_start:
            return
        self.counter.increment(1)

    def clear(self) -> None:
        """
        清零
        Returns:

        """
        # 清除超时、计数
        self.timer.clear()
        self.counter.clear()

    def start(self) -> None:
        if not self.is_start:
            self.__states = RunStates.START
            # 获取开始时间
            self._set_st(self.timer.get_current())

    def restart(self) -> None:
        """
        重新开始
        Returns:

        """
        self.clear()
        self._set_st(self.timer.get_current())

    def stop(self) -> None:
        self.__states = RunStates.STOP

    def _output_log(self) -> None:
        logger.warning(f"current qps: {self.counter.count}/s")

    def _set_st(self, current: NumberTypes) -> None:
        """
        更新开始时间
        Args:
            current:

        Returns:

        """
        if not isinstance(current, (int, float)):
            return
        self.__st = current

    def _set_timeout(self) -> None:
        """
        设置超时
        Returns:

        """
        if self.timer.get_current() - self.__st >= self.__timeout:
            self.timer.set_timeout()

    def __call__(self, *args, **kwargs) -> None:
        self.run()

    def run(self):
        self.start()
        self.auto_increment()
        self._set_timeout()

        if self.timer.is_timeout:
            self._output_log()
            self.restart()



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from w8_auto_py.typings import NumberTypes
from w8_auto_py.global_vars import TimeUnit
from w8_auto_py.common.qps import QueryPreSecond


def create_qps(timeout: NumberTypes,
               date_unit: TimeUnit = TimeUnit.MS) -> QueryPreSecond:
    """
    创建 qps
    Args:
        timeout:
        date_unit:

    Returns:

    """
    return QueryPreSecond(timeout, date_unit)



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import asyncio
import typing
from collections import deque, namedtuple

from w8_auto_py.common.emitter import AbstractEmitterTask
from w8_auto_py.common.validator import is_async_task

ATask = asyncio.Task
ATaskDeque = typing.Deque[ATask]
ATaskIterable = typing.Iterable[ATask]
JoinReturnType = typing.TypeVar("JoinReturnType")

Done = namedtuple("Done", ["index", "value"])

async def join(task: AbstractEmitterTask, **kwargs) -> typing.List[Done]:
    done = []
    if not isinstance(task, AbstractEmitterTask):
        raise TypeError("task params type error")

    for index, value in enumerate(asyncio.as_completed(task)):
        result = await value
        done.append(Done(index, result))

    return done


class EmitterQueue(AbstractEmitterTask):
    """
    使用队列实现
    """

    def __init__(self, max_len: int = None):
        self.__max_len = max_len
        self.__queue = deque(maxlen=self.max_len)

    @property
    def empty(self) -> bool:
        return self.length <= 0

    @property
    def max_len(self) -> int:
        if not isinstance(self.__max_len, int):
            self.__max_len = 2 ** 31
        return abs(self.__max_len)

    @property
    def length(self):
        return len(self.queue)

    @property
    def queue(self) -> ATaskDeque:
        return self.__queue

    def add_task(self, task: ATask, **kwargs) -> None:
        """
        添加任务
        :param task:
        :param kwargs:
        :return:
        """
        if not is_async_task(task):
            raise TypeError("task must be asyncio.Task, use asyncio.create_task()")

        self.queue.append(task)

    def __iter__(self) -> ATaskIterable:
        for _ in enumerate(range(self.length)):
            yield self.popleft()

    def popleft(self) -> ATask:
        if self.length:
            return self.queue.popleft()



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import time
import enum

from w8_auto_py.typings import NumberTypes
from w8_auto_py.global_vars import TimeUnit
from w8_auto_py.common.qps import AbstractQpsTimer
from w8_auto_py.util import EnumUtil


class TimeoutStates(enum.Enum):
    """
    超时状态
    """
    TIMEOUT = 1
    UN_TIMEOUT = -1


class QpsTimer(AbstractQpsTimer):

    def __init__(self, date_unit: TimeUnit):
        self.__date_unit = date_unit
        self.__cache = None
        self.__state: TimeoutStates = TimeoutStates.UN_TIMEOUT

    @property
    def is_timeout(self) -> bool:
        return self.__state == TimeoutStates.TIMEOUT

    @property
    def date_unit(self) -> NumberTypes:
        if not self.__cache:
            self.__cache = EnumUtil.get_enum_value(self.__date_unit)
        return self.__cache

    def get_current(self) -> NumberTypes:
        """
        获取当前时间
        Returns:

        """
        return int(round(time.time() / self.date_unit))

    def set_timeout(self) -> None:
        """
        设置超时
        Returns:

        """
        self.__state = TimeoutStates.TIMEOUT

    def clear(self) -> None:
        """
        清除超时
        Returns:

        """
        self.__state = TimeoutStates.UN_TIMEOUT



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import asyncio

from w8_auto_py.typings import ArgsType, Function, KwargsType, PoolExecutorTypes
from w8_auto_py import NotCallableError
from w8_auto_py.common.validator import callable_validator
from w8_auto_py.common.runner import AbstractRunner
from w8_auto_py.common.runner import create_runner_strategy, RunStrategy


class AsyncRunner(AbstractRunner):

    def __init__(self,
                 coro_or_func: Function,
                 args: ArgsType = None,
                 kwargs: KwargsType = None,
                 executor: PoolExecutorTypes = None):
        """

        Args:
            coro_or_func:   函数、协程函数、协程对象
            args:           函数或协程函数 位置参数
            kwargs:         函数或协程函数 关键字参数
            executor:       普通函数的执行器，默认线程，支持进程
        """

        self.__coro_or_func = coro_or_func
        self.__args = args
        self.__kwargs = kwargs
        self.__executor = executor

    @property
    def coro_or_func(self) -> Function:
        return self.__coro_or_func

    @property
    def args(self) -> ArgsType:
        if not isinstance(self.__args, tuple):
            self.__args = ()
        return self.__args

    @property
    def kwargs(self) -> KwargsType:
        if not isinstance(self.__kwargs, dict):
            self.__kwargs = {}
        return self.__kwargs

    @property
    def executor(self) -> PoolExecutorTypes:
        return self.__executor

    def _is_callable(self, func: Function) -> bool:
        try:
            return callable_validator(f"{func} 需要一个函数或对象").is_validate(func)
        except NotCallableError:
            return False

    def _is_coroutine_function(self, func: Function) -> bool:
        return asyncio.iscoroutinefunction(func)

    def _is_coroutine(self, func: Function) -> bool:
        return asyncio.iscoroutine(func)

    async def _function_strategy(self, runner, executor: PoolExecutorTypes = None):
        """
        普通函数执行策略
        Args:
            runner:
            executor:

        Returns:

        """
        strategy = create_runner_strategy(
            strategy=RunStrategy.FUNCTION,
            runner=runner,
            executor=executor
        )
        return await strategy.execute()

    async def _coro_function_strategy(self, runner, kwargs):
        """
        协程函数执行策略
        Args:
            runner:
            kwargs:

        Returns:

        """

        strategy = create_runner_strategy(
            strategy=RunStrategy.COROUTINE_FUNCTION,
            runner=runner, kwargs=kwargs
        )
        return await strategy.execute()

    async def _coro_strategy(self, runner):
        """
        协程对象策略
        Args:
            runner:

        Returns:

        """

        strategy = create_runner_strategy(
            strategy=RunStrategy.COROUTINE,
            runner=runner
        )

        return await strategy.execute()

    async def _impl_run(self, runner):

        if self._is_coroutine(runner.coro_or_func):
            # print("coroutine strategy ...")
            return await self._coro_strategy(runner)

        elif self._is_coroutine_function(runner.coro_or_func):
            # print("coroutine function strategy ...")
            return await self._coro_function_strategy(runner, runner.kwargs)

        elif self._is_callable(runner.coro_or_func):
            # print("function strategy ...")
            return await self._function_strategy(runner, runner.executor)

        else:
            # print("not support strategy ...")
            raise ValueError("coro_or_func 参数非法")

    def run(self):
        return asyncio.run(self._impl_run(self))



# @Time     : 2021/3/28
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import asyncio
import enum
import functools
import typing

from w8_auto_py.typings import ArgsType, KwargsType, PoolExecutorTypes
from w8_auto_py.util import EnumUtil
from w8_auto_py.common.emitter import AbstractEmitterStrategy, Done, IEmitter, join


class CoroutineEmitterStrategy(AbstractEmitterStrategy):
    """
    协程函数策略
    """

    def __init__(self,
                 emitter: IEmitter,
                 args: ArgsType = None,
                 kwargs: KwargsType = None):
        self.__emitter = emitter
        self.__args = () if not args else args
        self.__kwargs = {} if not kwargs else kwargs

    @property
    def emitter(self) -> IEmitter:
        return self.__emitter

    @property
    def args(self) -> ArgsType:
        return self.__args

    @property
    def kwargs(self) -> KwargsType:
        return self.__kwargs

    async def _impl_run(self, emitter: IEmitter, args: ArgsType, kwargs: KwargsType):
        for _ in enumerate(range(emitter.max_len)):
            t = asyncio.create_task(
                emitter.coro_or_func(*args, **kwargs)
            )
            emitter.task.add_task(t)
            await asyncio.sleep(emitter.delay)

        return await join(emitter.task)

    async def run(self) -> typing.List[Done]:
        return await self._impl_run(self.emitter, self.args, self.kwargs)


class FunctionEmitterStrategy(AbstractEmitterStrategy):
    """
    普通函数策略
    """

    def __init__(self,
                 emitter: IEmitter,
                 args: ArgsType = None,
                 executor: PoolExecutorTypes = None):
        self.__emitter = emitter
        self.__args = () if not args else args
        self.__executor = executor

    @property
    def emitter(self) -> IEmitter:
        return self.__emitter

    @property
    def args(self) -> ArgsType:
        return self.__args

    @property
    def executor(self) -> PoolExecutorTypes:
        return self.__executor

    async def _create_async_func(self,
                                 executor: PoolExecutorTypes,
                                 emitter: IEmitter,
                                 args: ArgsType):
        loop = asyncio.get_running_loop()
        if not loop:
            loop = asyncio.get_event_loop()

        return await loop.run_in_executor(executor, emitter.coro_or_func, *args)

    async def _impl_run(self,
                        executor: PoolExecutorTypes,
                        emitter: IEmitter,
                        args: ArgsType) -> typing.List[Done]:

        for _ in enumerate(range(emitter.max_len)):
            t = asyncio.create_task(
                self._create_async_func(executor, emitter, args)
            )
            emitter.task.add_task(t)
            await asyncio.sleep(emitter.delay)

        return await join(emitter.task)

    async def run(self) -> typing.List[Done]:
        return await self._impl_run(self.executor, self.emitter, self.args)


class EmitterStrategy(enum.Enum):
    COROUTINE = CoroutineEmitterStrategy
    FUNCTION = FunctionEmitterStrategy


def create_emitter_strategy(strategy: EmitterStrategy,
                            emitter: IEmitter,
                            **kwargs) -> AbstractEmitterStrategy:
    klass = EnumUtil.get_enum_value(strategy)
    f = functools.partial(klass, emitter=emitter)
    return f(**kwargs)



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import asyncio
import enum
import functools

from w8_auto_py.typings import Function, KwargsType, PoolExecutorTypes
from w8_auto_py.common.runner import AbstractRunStrategy, IRunner
from w8_auto_py.common.validator import callable_validator, coroutine_func_validator
from w8_auto_py.util import EnumUtil


class CoroutineStrategy(AbstractRunStrategy):
    """
    协程对象策略
    """

    def __init__(self, runner: IRunner):
        self.__runner = runner

    @property
    def runner(self):
        return self.__runner

    async def execute(self):
        return await self.runner.coro_or_func


class CoroutineFunctionStrategy(AbstractRunStrategy):
    """
    协程函数策略
    """

    def __init__(self, runner: IRunner, kwargs: KwargsType = None):
        self.__runner = runner
        self.__kwargs = kwargs

    @property
    def runner(self) -> IRunner:
        return self.__runner

    @property
    def kwargs(self) -> KwargsType:
        return self.__kwargs

    def _is_coroutine_func(self, func: Function) -> bool:
        return coroutine_func_validator(f"{func} 不是一个协程函数").is_validate(func)

    async def _impl_execute(self, runner: IRunner, kwargs):
        """
        实现
        :param runner:
        :param kwargs:
        :return:
        """
        return await runner.coro_or_func(*runner.args, **kwargs)

    async def execute(self):
        return await self._impl_execute(self.runner,
                                        kwargs=self.kwargs)


class FunctionStrategy(AbstractRunStrategy):
    """
    普通函数策略
    """

    def __init__(self,
                 runner: IRunner,
                 executor: PoolExecutorTypes = None):
        self.__runner = runner
        self.__executor = executor

    @property
    def runner(self) -> IRunner:
        return self.__runner

    @property
    def executor(self) -> PoolExecutorTypes:
        return self.__executor

    def _is_callable(self, func: Function) -> bool:
        return callable_validator(f"{func} 不是一个可调用对象").is_validate(func)

    async def _impl_execute(self,
                            runner: IRunner,
                            executor: PoolExecutorTypes = None):
        """
        execute 实现
        :param runner
        :param executor:
        :return:
        """
        loop = asyncio.get_running_loop()
        if not loop:
            loop = asyncio.get_event_loop()

        fut = loop.run_in_executor(executor, runner.coro_or_func, *runner.args)
        return await fut

    async def execute(self):
        return await self._impl_execute(self.runner,
                                        executor=self.executor)


class RunStrategy(enum.Enum):
    COROUTINE = CoroutineStrategy
    COROUTINE_FUNCTION = CoroutineFunctionStrategy
    FUNCTION = FunctionStrategy


def create_runner_strategy(strategy: RunStrategy,
                           runner: IRunner,
                           **kwargs) -> AbstractRunStrategy:
    if not isinstance(strategy, RunStrategy):
        strategy = RunStrategy.COROUTINE_FUNCTION

    f = functools.partial(
        EnumUtil.get_enum_value(strategy),
        runner=runner
    )
    return f(**kwargs)



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import typing

from w8_auto_py.typings import Function
from w8_auto_py import (
    NotCallableError,
    NotCoroutineError,
    NotCoroutineFunctionError
)
from w8_auto_py.common.validator import AbstractValidator
from w8_auto_py.common.validator import is_function, is_coroutine, is_coroutine_function


class CallableValidator(AbstractValidator):

    def is_validate(self, value: Function) -> bool:
        if not is_function(value):
            raise NotCallableError(f"{self.error_message}, {value}")
        return True


class CoroutineValidator(AbstractValidator):

    def is_validate(self, value: typing.Coroutine) -> bool:
        if not is_coroutine(value):
            raise NotCoroutineError(f"{self.error_message}, {value}")
        return True


class CoroutineFunctionValidator(AbstractValidator):

    def is_validate(self, value: Function) -> bool:
        if not is_coroutine_function(value):
            raise NotCoroutineFunctionError(f"{self.error_message}, {value}")
        return True



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import abc
import typing

from w8_auto_py.typings import ArgsType, Interface

T = typing.TypeVar("T")
RunnerReturnType = typing.TypeVar("RunnerReturnType")


class AbstractRunner(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def coro_or_func(self) -> T:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def args(self) -> ArgsType:
        raise NotImplementedError()

    @abc.abstractmethod
    def run(self) -> RunnerReturnType:
        raise NotImplementedError("NotImplementedError .run()")


class AbstractRunStrategy(metaclass=abc.ABCMeta):
    """
    策略接口
    """

    @property
    @abc.abstractmethod
    def runner(self):
        raise NotImplementedError("NotImplementedError .execute()")

    @abc.abstractmethod
    async def execute(self) -> RunnerReturnType:
        raise NotImplementedError("NotImplementedError .execute()")


class IRunner(Interface):
    """
    runner 接口
    """

    @property
    def coro_or_func(self) -> T:
        raise NotImplementedError()

    @property
    def args(self) -> ArgsType:
        raise NotImplementedError()

    def run(self) -> T:
        raise NotImplementedError()


class IRunStrategy(Interface):
    """
    RunStrategy 接口
    """

    @property
    def runner(self) -> IRunner:
        raise NotImplementedError()

    async def execute(self) -> RunnerReturnType:
        raise NotImplementedError()



# @Time     : 2021/3/25
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import abc
from w8_auto_py.typings import Function, FunctionIterable, FunctionSet


class AbstractCallback(metaclass=abc.ABCMeta):

    def __init__(self, *callbacks):
        self._set: FunctionSet = set(callbacks)

    @property
    def set(self) -> FunctionSet:
        return self._set

    @property
    def empty(self) -> bool:
        return not self.length

    @property
    def length(self) -> int:
        return len(self.set)

    def remove(self, callback: Function) -> None:
        if self.empty:
            return None

        return self.set.remove(callback)

    def __iter__(self) -> FunctionIterable:
        if not self.empty:
            for callback in self.set:
                yield callback

    @abc.abstractmethod
    def add(self, callback: Function) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def extend(self, *callbacks) -> None:
        raise NotImplementedError()



# @Time     : 2021/3/27
# @Project  : w8_auto_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from setuptools import setup, find_packages

_name = "w8_auto_py"
_author = "angel"
_author_email = "376355670@qq.com"
_description = "auto_tools"


def get_long_description():
    long_description = ""
    with open("README.md", mode="r", encoding="utf-8") as rd:
        long_description += rd.read()

    return long_description


setup(
    name=_name,
    version="0.2.2",
    author=_author,
    author_email=_author_email,
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7"
)



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import abc
import typing
from collections.abc import Awaitable

from w8_auto_py.typings import Function, Interface, NumberTypes
from w8_auto_py.global_vars import TimeUnit

T = typing.TypeVar("T")
EmitterReturnType = typing.TypeVar("EmitterReturnType")
EmitterMasterReturnType = typing.TypeVar("EmitterMasterReturnType")


def delay(delay: NumberTypes, time_unit: TimeUnit = TimeUnit.MS) -> NumberTypes:
    if not isinstance(delay, int):
        delay = 1

    if not isinstance(time_unit, TimeUnit):
        time_unit = TimeUnit.MS

    return delay * time_unit.value


class AbstractEmitterTask(metaclass=abc.ABCMeta):
    """
    抽象并发器任务
    """

    @abc.abstractmethod
    def add_task(self, task: T, **kwargs) -> None:
        """
        添加
        :param task:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def length(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def empty(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def __iter__(self) -> typing.Iterable[T]:
        raise NotImplementedError()


class AbstractEmitter(Awaitable, metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def coro_or_func(self) -> Function:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def delay(self) -> NumberTypes:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def max_len(self) -> int:
        raise NotImplementedError()

    @property
    def task(self) -> AbstractEmitterTask:
        raise NotImplementedError()

    @abc.abstractmethod
    async def run(self, *args, **kwargs) -> T:
        raise NotImplementedError()

    async def start(self) -> T:
        return await self.run()

    def __await__(self):
        return self.start().__await__()


class AbstractEmitterMaster(Awaitable, metaclass=abc.ABCMeta):
    """
    抽象并发管理组
    """

    def __init__(self,
                 task: AbstractEmitterTask,
                 delay: NumberTypes,
                 time_unit: TimeUnit = TimeUnit.MS):
        self.__task = task
        self.__delay = delay
        self.__time_unit = time_unit

        self.__cache_delay: NumberTypes = 0

    @property
    def delay(self) -> NumberTypes:
        if not self.__cache_delay:
            self.__cache_delay = delay(self.__delay, self.__time_unit)
        return self.__cache_delay

    @property
    def task(self) -> AbstractEmitterTask:
        return self.__task

    @abc.abstractmethod
    def add(self, emitter: AbstractEmitter, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def extends(self, *emitters, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def run(self, *args, **kwargs) -> EmitterMasterReturnType:
        raise NotImplementedError()

    async def start(self, *args, **kwargs) -> EmitterMasterReturnType:
        return await self.run(*args, **kwargs)

    def __await__(self):
        return self.start().__await__()


class AbstractEmitterStrategy(metaclass=abc.ABCMeta):
    """
    抽象并发策略
    """

    @property
    @abc.abstractmethod
    def emitter(self) -> AbstractEmitter:
        raise NotImplementedError()

    @abc.abstractmethod
    async def run(self, *args, **kwargs) -> EmitterReturnType:
        raise NotImplementedError()


class IEmitterTask(Interface):
    """
    并发任务接口
    """

    def add_task(self, task: T, **kwargs) -> None:
        raise NotImplementedError()

    def __iter__(self) -> typing.Iterable[T]:
        raise NotImplementedError()


class IEmitter(Interface):
    """
    并发器接口
    """

    @property
    def coro_or_func(self) -> Function:
        raise NotImplementedError()

    @property
    def delay(self) -> NumberTypes:
        raise NotImplementedError()

    @property
    def max_len(self) -> int:
        raise NotImplementedError()

    @property
    def task(self) -> IEmitterTask:
        raise NotImplementedError()

    async def start(self) -> EmitterReturnType:
        raise NotImplementedError()


class IEmitterMaster(Interface):

    @abc.abstractmethod
    def add(self, emitter: AbstractEmitter, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def extends(self, *emitters, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def start(self) -> EmitterMasterReturnType:
        raise NotImplementedError()

    @abc.abstractmethod
    async def join(self) -> EmitterMasterReturnType:
        raise NotImplementedError()



# @Time     : 2021/3/26
# @Project  : w8_auto_utils
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import enum
import functools

from w8_auto_py.typings import ArgsType, Function, KwargsType, NumberTypes, PoolExecutorTypes
from w8_auto_py.global_vars import TimeUnit
from w8_auto_py.common.emitter import (
    AbstractEmitter,
    Emitter,
    EmitterCollection,
    EmitterGroup,
    EmitterReturnType,
    EmitterQueue
)
from w8_auto_py.common.runner import create_arunner, run
from w8_auto_py.util import EnumUtil


class Emitters(enum.Enum):
    EMITTER = Emitter
    EMITTER_GROUP = EmitterGroup
    EMITTER_COLLECTION = EmitterCollection


class EmitterFactory:

    @classmethod
    def create_task(cls, max_len: int) -> EmitterQueue:
        return EmitterQueue(max_len=max_len)

    @classmethod
    def _factory(cls,
                 emitters: Emitters,
                 delay: NumberTypes = 500,
                 time_unit: TimeUnit = TimeUnit.MS,
                 **kwargs) -> AbstractEmitter:
        klass = EnumUtil.get_enum_value(emitters)
        f = functools.partial(
            klass,
            delay=delay,
            time_unit=time_unit
        )
        return f(**kwargs)

    @classmethod
    def create(cls,
               emitters: Emitters,
               task: EmitterQueue,
               delay: NumberTypes = 500,
               time_unit: TimeUnit = TimeUnit.MS,
               **kwargs) -> AbstractEmitter:
        return cls._factory(
            emitters=emitters,
            task=task,
            delay=delay,
            time_unit=time_unit,
            **kwargs
        )


def create_emitter(coro_or_func: Function,
                   args: ArgsType = None,
                   kwargs: KwargsType = None,
                   executor: PoolExecutorTypes = None,
                   delay: NumberTypes = 500,
                   max_len: int = None,
                   time_unit: TimeUnit = TimeUnit.MS) -> Emitter:
    """
    创建并发器
    :param coro_or_func:
    :param args:
    :param kwargs:
    :param executor:
    :param delay:
    :param max_len:
    :param time_unit:
    :return:
    """

    task = EmitterFactory.create_task(max_len)
    return EmitterFactory.create(
        emitters=Emitters.EMITTER,
        coro_or_func=coro_or_func,
        task=task,
        delay=delay,
        max_len=max_len,
        time_unit=time_unit,
        args=args,
        kwargs=kwargs,
        executor=executor
    )


def emitter_decorator(max_len: int = None,
                      delay: NumberTypes = 500,
                      time_unit: TimeUnit = TimeUnit.MS,
                      executor: PoolExecutorTypes = None) -> Emitter:
    """
    并发包装器
    :param max_len:
    :param delay:
    :param time_unit:
    :param executor:
    :return:
    """

    def _decorate(coro_or_func: Function):
        @functools.wraps(coro_or_func)
        def _wrapper(*args, **kwargs) -> Emitter:
            return create_emitter(
                coro_or_func=coro_or_func,
                delay=delay,
                max_len=max_len,
                time_unit=time_unit,
                args=args,
                kwargs=kwargs,
                executor=executor
            )

        return _wrapper

    return _decorate


def create_emitter_group(max_len: int = None,
                         delay: NumberTypes = 500,
                         time_unit: TimeUnit = TimeUnit.MS) -> EmitterGroup:
    """
    创建 并发组
    Args:
        max_len:
        delay:
        time_unit:

    Returns:

    """
    task = EmitterFactory.create_task(max_len)
    return EmitterFactory.create(
        emitters=Emitters.EMITTER_GROUP,
        task=task,
        delay=delay,
        time_unit=time_unit
    )


def create_emitter_collection(max_Len: int = None,
                              delay: NumberTypes = 500,
                              time_unit: TimeUnit = TimeUnit.MS) -> EmitterCollection:
    """
    创建 并发集合
    Args:
        max_Len:
        delay:
        time_unit:

    Returns:

    """
    task = EmitterFactory.create_task(max_Len)
    return EmitterFactory.create(
        emitters=Emitters.EMITTER_COLLECTION,
        task=task,
        delay=delay,
        time_unit=time_unit
    )


def start(emitter: AbstractEmitter) -> EmitterReturnType:
    """
    运行并发器
    Args:
        emitter:

    Returns:

    """
    runner = create_arunner(emitter.start())
    return run(runner)
